
#import "EAFRead.h"

// *** typical block size used by CA on my machine
#define NUM_FRAMES_PER_READ   512

@implementation EAFRead

//-----------------------------------------------------------------------------------------------------------------------------------------------------------

- (OSStatus) openFileForRead:(NSURL*)fileURL sr:(Float64)sampleRate channels:(int)numChannels
{
	OSStatus err = noErr;
	FSRef theRef;
	UInt32 propSize;		
	
	extAFReachedEOF = NO;
	
	if(!CFURLGetFSRef((CFURLRef)fileURL, &theRef)) return -1;
	
	err = ExtAudioFileOpen(&theRef, &extAFRef);
	if (err) {NSLog(@"!!! Error in ExtAudioFileOpen, %d", err); return err;}
	
	AudioStreamBasicDescription fileFormat;
	propSize = sizeof(fileFormat);
   memset(&fileFormat, 0, sizeof(AudioStreamBasicDescription));
	
	err = ExtAudioFileGetProperty(extAFRef, kExtAudioFileProperty_FileDataFormat, &propSize, &fileFormat);
	if (err) {NSLog(@"!!! Error in ExtAudioFileGetProperty, %d", err); return err;}
	
	extAFRateRatio = sampleRate / fileFormat.mSampleRate;
   NSLog(@"Converting sample rate from %f to %f", fileFormat.mSampleRate, sampleRate);
   
	AudioStreamBasicDescription clientFormat;
	propSize = sizeof(clientFormat);
   memset(&clientFormat, 0, sizeof(AudioStreamBasicDescription));
	clientFormat.mFormatID				= kAudioFormatLinearPCM;
	clientFormat.mSampleRate			= sampleRate;
   clientFormat.mFormatFlags        = kAudioFormatFlagIsFloat | kAudioFormatFlagIsPacked | kAudioFormatFlagsNativeEndian;
	clientFormat.mChannelsPerFrame   = numChannels;
	clientFormat.mBitsPerChannel     = sizeof(float) * 8;
	clientFormat.mFramesPerPacket    = 1;
	clientFormat.mBytesPerFrame      = clientFormat.mBitsPerChannel * clientFormat.mChannelsPerFrame / 8;
	clientFormat.mBytesPerPacket     = clientFormat.mFramesPerPacket * clientFormat.mBytesPerFrame;
   clientFormat.mReserved           = 0;
   
   err = ExtAudioFileSetProperty(extAFRef, kExtAudioFileProperty_ClientDataFormat, propSize, &clientFormat);
	if (err) {NSLog(@"!!! Error in ExtAudioFileSetProperty, %d", err); return err;}
	
	extAFNumChannels = numChannels;
	return err;
}

// ---------------------------------------------------------------------------------------------------------------------

- (OSStatus) closeFile
{
	OSStatus err = noErr;
	if (extAFRef) {
		ExtAudioFileDispose(extAFRef);
		extAFRef = nil;
	}
	return err;
}

// ---------------------------------------------------------------------------------------------------------------------

- (OSStatus) readFloatsConsecutive:(SInt64)numFrames intoArray:(float**)audio
{
	OSStatus err = noErr;
   
	if (!extAFRef)	return -1;
   	
	int kSegmentSize = (int)(numFrames * extAFNumChannels * extAFRateRatio + .5);
	if (extAFRateRatio < 1.) kSegmentSize = (int)(numFrames * extAFNumChannels / extAFRateRatio + .5);
   
	float *data = (float*)malloc(kSegmentSize*sizeof(float));
	if (!data) {
      NSLog(@"data is nil");
      goto error;
   }
   
	UInt32 numPackets = numFrames; // Frames to read
	UInt32 samples = numPackets * extAFNumChannels; // 2 channels (samples) per frame
   
	AudioBufferList bufList;
	bufList.mNumberBuffers = 1;
	bufList.mBuffers[0].mNumberChannels = extAFNumChannels; // Always 2 channels in this example
	bufList.mBuffers[0].mData = data; // data is a pointer (float*) to our sample buffer
	bufList.mBuffers[0].mDataByteSize = samples * sizeof(float);
	
	UInt32 loadedPackets = numPackets;
	
	err = ExtAudioFileRead(extAFRef, &loadedPackets, &bufList);
	if (err) goto error;
	
   if (audio) {
      for (long c = 0; c < extAFNumChannels; c++) {
         if (!audio[c]) continue;
         for (long v = 0; v < numFrames; v++) {
            if (v < loadedPackets) audio[c][v] = data[v*extAFNumChannels+c];
            else audio[c][v] = 0.;
         }
      }
   }
   
error:
	free(data);
	if (err != noErr) return err;
	if (loadedPackets < numFrames) extAFReachedEOF = YES;
	return loadedPackets;
}

// ---------------------------------------------------------------------------------

-(id)init
{
   
   // This is for quick'n dirty testing so I'm just sticking all the code into init
   
   extAFRateRatio = 1.;
	extAFRef=nil;
	extAFNumChannels = 0;
	extAFReachedEOF = NO;

   NSOpenPanel *panel = [NSOpenPanel openPanel];
	
	[panel setResolvesAliases:YES];
	[panel setAllowsMultipleSelection:NO];
	[panel setPrompt: @"Open sound file"];
	[panel setMessage: @"We're trying to crash ExtAudioFileRead! Try opening an MP3 file that is at least 30 seconds long."];
	
	int result = [panel runModalForTypes:[NSSound soundUnfilteredFileTypes]];
	
	if (result == NSOKButton) {
   
      if ([self openFileForRead:[panel URL] sr:44100 channels:2] < 0) {
         NSLog(@"Cannot open file - exiting");
         goto error;
      }
      
      long frames = NUM_FRAMES_PER_READ;
      unsigned long framesRead=0;
      long previousTimeSeconds = -1;
      for (;;) {
         long timeSeconds = (long)(framesRead/44100.);
         if (previousTimeSeconds != timeSeconds) {
            NSLog(@"Reading %d frames, seconds: %d", frames, timeSeconds);
            previousTimeSeconds = timeSeconds;
         }
         if ([self readFloatsConsecutive:frames intoArray:nil] < 0) {   // we're not interested in the actual audio data, all we want to do is read from the file
            NSLog(@"Cannot read file - exiting");
            break;
         }
         framesRead += frames;
         if (extAFReachedEOF) break;
      }
      [self closeFile];
      NSLog(@"Done. No crash - good!");
   }
error:
   return self;   
}

// ---------------------------------------------------------------------------------


@end
